@@ -6,7 +6,7 @@ Discord API Wrapper
 
 A basic wrapper for the Discord API.
 
-:copyright: (c) 2015-2016 Rapptz
+:copyright: (c) 2015-2019 Rapptz
 :license: MIT, see LICENSE for more details.
 
 """
@@ -14,36 +14,46 @@ A basic wrapper for the Discord API.
 __title__ = 'discord'
 __author__ = 'Rapptz'
 __license__ = 'MIT'
-__copyright__ = 'Copyright 2015-2016 Rapptz'
-__version__ = '0.16.12'
+__copyright__ = 'Copyright 2015-2019 Rapptz'
+__version__ = '1.0.0'
 
-from .client import Client, AppInfo, ChannelPermissions
-from .user import User
-from .game import Game
-from .emoji import Emoji
-from .channel import Channel, PrivateChannel
-from .server import Server
+from collections import namedtuple
+import logging
+
+from .client import Client
+from .appinfo import AppInfo
+from .user import User, ClientUser, Profile
+from .emoji import Emoji, PartialEmoji
+from .activity import *
+from .channel import *
+from .guild import Guild
+from .relationship import Relationship
 from .member import Member, VoiceState
-from .message import Message
+from .message import Message, Attachment
+from .asset import Asset
 from .errors import *
 from .calls import CallMessage, GroupCall
 from .permissions import Permissions, PermissionOverwrite
 from .role import Role
+from .file import File
 from .colour import Color, Colour
-from .invite import Invite
+from .invite import Invite, PartialInviteChannel, PartialInviteGuild
+from .widget import Widget, WidgetMember, WidgetChannel
 from .object import Object
 from .reaction import Reaction
-from . import utils, opus, compat
-from .voice_client import VoiceClient
-from .enums import ChannelType, ServerRegion, Status, MessageType, VerificationLevel
-from collections import namedtuple
+from . import utils, opus, abc
+from .enums import *
 from .embeds import Embed
-
-import logging
+from .shard import AutoShardedClient
+from .player import *
+from .webhook import *
+from .voice_client import VoiceClient
+from .audit_logs import AuditLogChanges, AuditLogEntry, AuditLogDiff
+from .raw_models import *
 
 VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
 
-version_info = VersionInfo(major=0, minor=16, micro=12, releaselevel='final', serial=0)
+version_info = VersionInfo(major=1, minor=0, micro=0, releaselevel='final', serial=0)
 
 try:
     from logging import NullHandler
