@@ -3,7 +3,7 @@
 """
 The MIT License (MIT)
 
-Copyright (c) 2015-2016 Rapptz
+Copyright (c) 2015-2019 Rapptz
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
@@ -24,36 +24,35 @@ FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 """
 
-import aiohttp
 import asyncio
 import json
-import sys
 import logging
+import sys
+from urllib.parse import quote as _uriquote
 import weakref
-import datetime
-from email.utils import parsedate_to_datetime
 
-log = logging.getLogger(__name__)
+import aiohttp
 
 from .errors import HTTPException, Forbidden, NotFound, LoginFailure, GatewayNotFound
 from . import __version__, utils
 
-@asyncio.coroutine
-def json_or_text(response):
-    text = yield from response.text(encoding='utf-8')
+log = logging.getLogger(__name__)
+
+async def json_or_text(response):
+    text = await response.text(encoding='utf-8')
     if response.headers['content-type'] == 'application/json':
         return json.loads(text)
     return text
 
 class Route:
-    BASE = 'https://discordapp.com/api/v6'
+    BASE = 'https://discordapp.com/api/v7'
 
     def __init__(self, method, path, **parameters):
         self.path = path
         self.method = method
         url = (self.BASE + self.path)
         if parameters:
-            self.url = url.format(**parameters)
+            self.url = url.format(**{k: _uriquote(v) if isinstance(v, str) else v for k, v in parameters.items()})
         else:
             self.url = url
 
@@ -87,21 +86,26 @@ class HTTPClient:
     SUCCESS_LOG = '{method} {url} has received {text}'
     REQUEST_LOG = '{method} {url} with {json} has returned {status}'
 
-    def __init__(self, connector=None, *, loop=None):
+    def __init__(self, connector=None, *, proxy=None, proxy_auth=None, loop=None):
         self.loop = asyncio.get_event_loop() if loop is None else loop
         self.connector = connector
-        self.session = aiohttp.ClientSession(connector=connector, loop=self.loop)
+        self.__session = None # filled in static_login
         self._locks = weakref.WeakValueDictionary()
         self._global_over = asyncio.Event(loop=self.loop)
         self._global_over.set()
         self.token = None
         self.bot_token = False
+        self.proxy = proxy
+        self.proxy_auth = proxy_auth
 
         user_agent = 'DiscordBot (https://github.com/Rapptz/discord.py {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
         self.user_agent = user_agent.format(__version__, sys.version_info, aiohttp.__version__)
 
-    @asyncio.coroutine
-    def request(self, route, *, header_bypass_delay=None, **kwargs):
+    def recreate(self):
+        if self.__session.closed:
+            self.__session = aiohttp.ClientSession(connector=self.connector, loop=self.loop)
+
+    async def request(self, route, *, files=None, header_bypass_delay=None, **kwargs):
         bucket = route.bucket
         method = route.method
         url = route.url
@@ -119,63 +123,81 @@ class HTTPClient:
 
         if self.token is not None:
             headers['Authorization'] = 'Bot ' + self.token if self.bot_token else self.token
-
         # some checking if it's a JSON request
         if 'json' in kwargs:
             headers['Content-Type'] = 'application/json'
             kwargs['data'] = utils.to_json(kwargs.pop('json'))
 
+        try:
+            reason = kwargs.pop('reason')
+        except KeyError:
+            pass
+        else:
+            if reason:
+                headers['X-Audit-Log-Reason'] = _uriquote(reason, safe='/ ')
+
         kwargs['headers'] = headers
 
+        # Proxy support
+        if self.proxy is not None:
+            kwargs['proxy'] = self.proxy
+        if self.proxy_auth is not None:
+            kwargs['proxy_auth'] = self.proxy_auth
+
         if not self._global_over.is_set():
             # wait until the global lock is complete
-            yield from self._global_over.wait()
+            await self._global_over.wait()
 
-        yield from lock
+        await lock.acquire()
         with MaybeUnlock(lock) as maybe_lock:
             for tries in range(5):
-                r = yield from self.session.request(method, url, **kwargs)
-                log.debug(self.REQUEST_LOG.format(method=method, url=url, status=r.status, json=kwargs.get('data')))
-                try:
+                if files:
+                    for f in files:
+                        f.reset(seek=tries)
+
+                async with self.__session.request(method, url, **kwargs) as r:
+                    log.debug('%s %s with %s has returned %s', method, url, kwargs.get('data'), r.status)
+
                     # even errors have text involved in them so this is safe to call
-                    data = yield from json_or_text(r)
+                    data = await json_or_text(r)
 
                     # check if we have rate limit header information
                     remaining = r.headers.get('X-Ratelimit-Remaining')
                     if remaining == '0' and r.status != 429:
                         # we've depleted our current bucket
                         if header_bypass_delay is None:
-                            now = parsedate_to_datetime(r.headers['Date'])
-                            reset = datetime.datetime.fromtimestamp(int(r.headers['X-Ratelimit-Reset']), datetime.timezone.utc)
-                            delta = (reset - now).total_seconds()
+                            delta = utils._parse_ratelimit_header(r)
                         else:
                             delta = header_bypass_delay
 
-                        fmt = 'A rate limit bucket has been exhausted (bucket: {bucket}, retry: {delta}).'
-                        log.info(fmt.format(bucket=bucket, delta=delta))
+                        log.debug('A rate limit bucket has been exhausted (bucket: %s, retry: %s).', bucket, delta)
                         maybe_lock.defer()
                         self.loop.call_later(delta, lock.release)
 
                     # the request was successful so just return the text/json
                     if 300 > r.status >= 200:
-                        log.debug(self.SUCCESS_LOG.format(method=method, url=url, text=data))
+                        log.debug('%s %s has received %s', method, url, data)
                         return data
 
                     # we are being rate limited
                     if r.status == 429:
-                        fmt = 'We are being rate limited. Retrying in {:.2} seconds. Handled under the bucket "{}"'
+                        if not isinstance(data, dict):
+                            # Banned by Cloudflare more than likely.
+                            raise HTTPException(r, data)
+
+                        fmt = 'We are being rate limited. Retrying in %.2f seconds. Handled under the bucket "%s"'
 
                         # sleep a bit
                         retry_after = data['retry_after'] / 1000.0
-                        log.info(fmt.format(retry_after, bucket))
+                        log.warning(fmt, retry_after, bucket)
 
                         # check if it's a global rate limit
                         is_global = data.get('global', False)
                         if is_global:
-                            log.info('Global rate limit has been hit. Retrying in {:.2} seconds.'.format(retry_after))
+                            log.warning('Global rate limit has been hit. Retrying in %.2f seconds.', retry_after)
                             self._global_over.clear()
 
-                        yield from asyncio.sleep(retry_after, loop=self.loop)
+                        await asyncio.sleep(retry_after, loop=self.loop)
                         log.debug('Done sleeping for the rate limit. Retrying...')
 
                         # release the global lock now that the
@@ -186,9 +208,9 @@ class HTTPClient:
 
                         continue
 
-                    # we've received a 502, unconditional retry
-                    if r.status == 502 and tries <= 5:
-                        yield from asyncio.sleep(1 + tries * 2, loop=self.loop)
+                    # we've received a 500 or 502, unconditional retry
+                    if r.status in {500, 502}:
+                        await asyncio.sleep(1 + tries * 2, loop=self.loop)
                         continue
 
                     # the usual error cases
@@ -198,75 +220,84 @@ class HTTPClient:
                         raise NotFound(r, data)
                     else:
                         raise HTTPException(r, data)
-                finally:
-                    # clean-up just in case
-                    yield from r.release()
 
-    def get(self, *args, **kwargs):
-        return self.request('GET', *args, **kwargs)
+            # We've run out of retries, raise.
+            raise HTTPException(r, data)
 
-    def put(self, *args, **kwargs):
-        return self.request('PUT', *args, **kwargs)
-
-    def patch(self, *args, **kwargs):
-        return self.request('PATCH', *args, **kwargs)
-
-    def delete(self, *args, **kwargs):
-        return self.request('DELETE', *args, **kwargs)
-
-    def post(self, *args, **kwargs):
-        return self.request('POST', *args, **kwargs)
+    async def get_from_cdn(self, url):
+        async with self.__session.get(url) as resp:
+            if resp.status == 200:
+                return await resp.read()
+            elif resp.status == 404:
+                raise NotFound(resp, 'asset not found')
+            elif resp.status == 403:
+                raise Forbidden(resp, 'cannot retrieve asset')
+            else:
+                raise HTTPException(resp, 'failed to get asset')
 
     # state management
 
-    @asyncio.coroutine
-    def close(self):
-        yield from self.session.close()
-
-    def recreate(self):
-        self.session = aiohttp.ClientSession(connector=self.connector, loop=self.loop)
+    async def close(self):
+        if self.__session:
+            await self.__session.close()
 
     def _token(self, token, *, bot=True):
         self.token = token
         self.bot_token = bot
+        self._ack_token = None
 
     # login management
 
-    @asyncio.coroutine
-    def email_login(self, email, password):
-        payload = {
-            'email': email,
-            'password': password
-        }
-
-        try:
-            data = yield from self.request(Route('POST', '/auth/login'), json=payload)
-        except HTTPException as e:
-            if e.response.status == 400:
-                raise LoginFailure('Improper credentials have been passed.') from e
-            raise
-
-        self._token(data['token'], bot=False)
-        return data
-
-    @asyncio.coroutine
-    def static_login(self, token, *, bot):
+    async def static_login(self, token, *, bot):
+        # Necessary to get aiohttp to stop complaining about session creation
+        self.__session = aiohttp.ClientSession(connector=self.connector, loop=self.loop)
         old_token, old_bot = self.token, self.bot_token
         self._token(token, bot=bot)
 
         try:
-            data = yield from self.request(Route('GET', '/users/@me'))
-        except HTTPException as e:
+            data = await self.request(Route('GET', '/users/@me'))
+        except HTTPException as exc:
             self._token(old_token, bot=old_bot)
-            if e.response.status == 401:
-                raise LoginFailure('Improper token has been passed.') from e
-            raise e
+            if exc.response.status == 401:
+                raise LoginFailure('Improper token has been passed.') from exc
+            raise
 
         return data
 
     def logout(self):
         return self.request(Route('POST', '/auth/logout'))
 
+    # Group functionality
+
+    def start_group(self, user_id, recipients):
+        payload = {
+            'recipients': recipients
+        }
+
+        return self.request(Route('POST', '/users/{user_id}/channels', user_id=user_id), json=payload)
+
+    def leave_group(self, channel_id):
+        return self.request(Route('DELETE', '/channels/{channel_id}', channel_id=channel_id))
+
+    def add_group_recipient(self, channel_id, user_id):
+        r = Route('PUT', '/channels/{channel_id}/recipients/{user_id}', channel_id=channel_id, user_id=user_id)
+        return self.request(r)
+
+    def remove_group_recipient(self, channel_id, user_id):
+        r = Route('DELETE', '/channels/{channel_id}/recipients/{user_id}', channel_id=channel_id, user_id=user_id)
+        return self.request(r)
+
+    def edit_group(self, channel_id, **options):
+        valid_keys = ('name', 'icon')
+        payload = {
+            k: v for k, v in options.items() if k in valid_keys
+        }
+
+        return self.request(Route('PATCH', '/channels/{channel_id}', channel_id=channel_id), json=payload)
+
+    def convert_group(self, channel_id):
+        return self.request(Route('POST', '/channels/{channel_id}/convert', channel_id=channel_id))
+
     # Message management
 
     def start_private_message(self, user_id):
@@ -276,9 +307,7 @@ class HTTPClient:
 
         return self.request(Route('POST', '/users/@me/channels'), json=payload)
 
-    # TODO: remove guild_id parameters here
-
-    def send_message(self, channel_id, content, *, guild_id=None, tts=False, embed=None):
+    def send_message(self, channel_id, content, *, tts=False, embed=None, nonce=None):
         r = Route('POST', '/channels/{channel_id}/messages', channel_id=channel_id)
         payload = {}
 
@@ -291,12 +320,15 @@ class HTTPClient:
         if embed:
             payload['embed'] = embed
 
+        if nonce:
+            payload['nonce'] = nonce
+
         return self.request(r, json=payload)
 
     def send_typing(self, channel_id):
         return self.request(Route('POST', '/channels/{channel_id}/typing', channel_id=channel_id))
 
-    def send_file(self, channel_id, buffer, *, guild_id=None, filename=None, content=None, tts=False, embed=None):
+    def send_files(self, channel_id, *, files, content=None, tts=False, embed=None, nonce=None):
         r = Route('POST', '/channels/{channel_id}/messages', channel_id=channel_id)
         form = aiohttp.FormData()
 
@@ -305,37 +337,42 @@ class HTTPClient:
             payload['content'] = content
         if embed:
             payload['embed'] = embed
+        if nonce:
+            payload['nonce'] = nonce
 
         form.add_field('payload_json', utils.to_json(payload))
-        form.add_field('file', buffer, filename=filename, content_type='application/octet-stream')
+        if len(files) == 1:
+            file = files[0]
+            form.add_field('file', file.fp, filename=file.filename, content_type='application/octet-stream')
+        else:
+            for index, file in enumerate(files):
+                form.add_field('file%s' % index, file.fp, filename=file.filename, content_type='application/octet-stream')
 
-        return self.request(r, data=form)
+        return self.request(r, data=form, files=files)
 
-    def delete_message(self, channel_id, message_id, guild_id=None):
-        r = Route('DELETE', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id,
-                                                                            message_id=message_id)
-        return self.request(r)
+    async def ack_message(self, channel_id, message_id):
+        r = Route('POST', '/channels/{channel_id}/messages/{message_id}/ack', channel_id=channel_id, message_id=message_id)
+        data = await self.request(r, json={'token': self._ack_token})
+        self._ack_token = data['token']
 
-    def delete_messages(self, channel_id, message_ids, guild_id=None):
+    def ack_guild(self, guild_id):
+        return self.request(Route('POST', '/guilds/{guild_id}/ack', guild_id=guild_id))
+
+    def delete_message(self, channel_id, message_id, *, reason=None):
+        r = Route('DELETE', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id, message_id=message_id)
+        return self.request(r, reason=reason)
+
+    def delete_messages(self, channel_id, message_ids, *, reason=None):
         r = Route('POST', '/channels/{channel_id}/messages/bulk_delete', channel_id=channel_id)
         payload = {
             'messages': message_ids
         }
 
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def edit_message(self, message_id, channel_id, content, *, guild_id=None, embed=None):
-        r = Route('PATCH', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id,
-                                                                           message_id=message_id)
-        payload = {}
-
-        if content:
-            payload['content'] = content
-
-        if embed:
-            payload['embed'] = embed
-
-        return self.request(r, json=payload)
+    def edit_message(self, message_id, channel_id, **fields):
+        r = Route('PATCH', '/channels/{channel_id}/messages/{message_id}', channel_id=channel_id, message_id=message_id)
+        return self.request(r, json=fields)
 
     def add_reaction(self, message_id, channel_id, emoji):
         r = Route('PUT', '/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me',
@@ -347,9 +384,14 @@ class HTTPClient:
                   channel_id=channel_id, message_id=message_id, member_id=member_id, emoji=emoji)
         return self.request(r, header_bypass_delay=0.25)
 
+    def remove_own_reaction(self, message_id, channel_id, emoji):
+        r = Route('DELETE', '/channels/{channel_id}/messages/{message_id}/reactions/{emoji}/@me',
+                  channel_id=channel_id, message_id=message_id, emoji=emoji)
+        return self.request(r, header_bypass_delay=0.25)
+
     def get_reaction_users(self, message_id, channel_id, emoji, limit, after=None):
         r = Route('GET', '/channels/{channel_id}/messages/{message_id}/reactions/{emoji}',
-                         channel_id=channel_id, message_id=message_id, emoji=emoji)
+                  channel_id=channel_id, message_id=message_id, emoji=emoji)
 
         params = {'limit': limit}
         if after:
@@ -371,44 +413,53 @@ class HTTPClient:
             'limit': limit
         }
 
-        if before:
+        if before is not None:
             params['before'] = before
-        if after:
+        if after is not None:
             params['after'] = after
-        if around:
+        if around is not None:
             params['around'] = around
 
         return self.request(Route('GET', '/channels/{channel_id}/messages', channel_id=channel_id), params=params)
 
     def pin_message(self, channel_id, message_id):
         return self.request(Route('PUT', '/channels/{channel_id}/pins/{message_id}',
-                            channel_id=channel_id, message_id=message_id))
+                                  channel_id=channel_id, message_id=message_id))
 
     def unpin_message(self, channel_id, message_id):
         return self.request(Route('DELETE', '/channels/{channel_id}/pins/{message_id}',
-                            channel_id=channel_id, message_id=message_id))
+                                  channel_id=channel_id, message_id=message_id))
 
     def pins_from(self, channel_id):
         return self.request(Route('GET', '/channels/{channel_id}/pins', channel_id=channel_id))
 
     # Member management
 
-    def kick(self, user_id, guild_id):
+    def kick(self, user_id, guild_id, reason=None):
         r = Route('DELETE', '/guilds/{guild_id}/members/{user_id}', guild_id=guild_id, user_id=user_id)
+        if reason:
+            # thanks aiohttp
+            r.url = '{0.url}?reason={1}'.format(r, _uriquote(reason))
+
         return self.request(r)
 
-    def ban(self, user_id, guild_id, delete_message_days=1):
+    def ban(self, user_id, guild_id, delete_message_days=1, reason=None):
         r = Route('PUT', '/guilds/{guild_id}/bans/{user_id}', guild_id=guild_id, user_id=user_id)
         params = {
-            'delete-message-days': delete_message_days
+            'delete-message-days': delete_message_days,
         }
+
+        if reason:
+            # thanks aiohttp
+            r.url = '{0.url}?reason={1}'.format(r, _uriquote(reason))
+
         return self.request(r, params=params)
 
-    def unban(self, user_id, guild_id):
+    def unban(self, user_id, guild_id, *, reason=None):
         r = Route('DELETE', '/guilds/{guild_id}/bans/{user_id}', guild_id=guild_id, user_id=user_id)
-        return self.request(r)
+        return self.request(r, reason=reason)
 
-    def server_voice_state(self, user_id, guild_id, *, mute=None, deafen=None):
+    def guild_voice_state(self, user_id, guild_id, *, mute=None, deafen=None, reason=None):
         r = Route('PATCH', '/guilds/{guild_id}/members/{user_id}', guild_id=guild_id, user_id=user_id)
         payload = {}
         if mute is not None:
@@ -417,7 +468,7 @@ class HTTPClient:
         if deafen is not None:
             payload['deaf'] = deafen
 
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
     def edit_profile(self, password, username, avatar, **fields):
         payload = {
@@ -434,60 +485,100 @@ class HTTPClient:
 
         return self.request(Route('PATCH', '/users/@me'), json=payload)
 
-    def change_my_nickname(self, guild_id, nickname):
+    def change_my_nickname(self, guild_id, nickname, *, reason=None):
+        r = Route('PATCH', '/guilds/{guild_id}/members/@me/nick', guild_id=guild_id)
         payload = {
             'nick': nickname
         }
-        return self.request(Route('PATCH', '/guilds/{guild_id}/members/@me/nick', guild_id=guild_id), json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def change_nickname(self, guild_id, user_id, nickname):
+    def change_nickname(self, guild_id, user_id, nickname, *, reason=None):
         r = Route('PATCH', '/guilds/{guild_id}/members/{user_id}', guild_id=guild_id, user_id=user_id)
         payload = {
             'nick': nickname
         }
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def edit_member(self, guild_id, user_id, **fields):
+    def edit_member(self, guild_id, user_id, *, reason=None, **fields):
         r = Route('PATCH', '/guilds/{guild_id}/members/{user_id}', guild_id=guild_id, user_id=user_id)
-        return self.request(r, json=fields)
+        return self.request(r, json=fields, reason=reason)
 
     # Channel management
 
-    def edit_channel(self, channel_id, **options):
-        valid_keys = ('name', 'topic', 'bitrate', 'user_limit', 'position')
+    def edit_channel(self, channel_id, *, reason=None, **options):
+        r = Route('PATCH', '/channels/{channel_id}', channel_id=channel_id)
+        valid_keys = ('name', 'parent_id', 'topic', 'bitrate', 'nsfw',
+                      'user_limit', 'position', 'permission_overwrites', 'rate_limit_per_user')
         payload = {
             k: v for k, v in options.items() if k in valid_keys
         }
 
-        return self.request(Route('PATCH', '/channels/{channel_id}', channel_id=channel_id), json=payload)
+        return self.request(r, reason=reason, json=payload)
 
-    def move_channel_position(self, guild_id, positions):
+    def bulk_channel_update(self, guild_id, data, *, reason=None):
         r = Route('PATCH', '/guilds/{guild_id}/channels', guild_id=guild_id)
-        return self.request(r, json=positions)
+        return self.request(r, json=data, reason=reason)
 
-    def create_channel(self, guild_id, name, channe_type, permission_overwrites=None):
+    def create_channel(self, guild_id, channel_type, *, reason=None, **options):
         payload = {
-            'name': name,
-            'type': channe_type
+            'type': channel_type
         }
 
-        if permission_overwrites is not None:
-            payload['permission_overwrites'] = permission_overwrites
+        valid_keys = ('name', 'parent_id', 'topic', 'bitrate', 'nsfw',
+                      'user_limit', 'position', 'permission_overwrites', 'rate_limit_per_user')
+        payload.update({
+            k: v for k, v in options.items() if k in valid_keys and v is not None
+        })
 
-        return self.request(Route('POST', '/guilds/{guild_id}/channels', guild_id=guild_id), json=payload)
+        return self.request(Route('POST', '/guilds/{guild_id}/channels', guild_id=guild_id), json=payload, reason=reason)
 
-    def delete_channel(self, channel_id):
-        return self.request(Route('DELETE', '/channels/{channel_id}', channel_id=channel_id))
+    def delete_channel(self, channel_id, *, reason=None):
+        return self.request(Route('DELETE', '/channels/{channel_id}', channel_id=channel_id), reason=reason)
+
+    # Webhook management
+
+    def create_webhook(self, channel_id, *, name, avatar=None):
+        payload = {
+            'name': name
+        }
+        if avatar is not None:
+            payload['avatar'] = avatar
+
+        return self.request(Route('POST', '/channels/{channel_id}/webhooks', channel_id=channel_id), json=payload)
+
+    def channel_webhooks(self, channel_id):
+        return self.request(Route('GET', '/channels/{channel_id}/webhooks', channel_id=channel_id))
+
+    def guild_webhooks(self, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/webhooks', guild_id=guild_id))
+
+    def get_webhook(self, webhook_id):
+        return self.request(Route('GET', '/webhooks/{webhook_id}', webhook_id=webhook_id))
 
     # Guild management
 
-    def leave_server(self, guild_id):
+    def get_guilds(self, limit, before=None, after=None):
+        params = {
+            'limit': limit
+        }
+
+        if before:
+            params['before'] = before
+        if after:
+            params['after'] = after
+
+        return self.request(Route('GET', '/users/@me/guilds'), params=params)
+
+    def leave_guild(self, guild_id):
         return self.request(Route('DELETE', '/users/@me/guilds/{guild_id}', guild_id=guild_id))
 
-    def delete_server(self, guild_id):
+    def get_guild(self, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}', guild_id=guild_id))
+
+    def delete_guild(self, guild_id):
         return self.request(Route('DELETE', '/guilds/{guild_id}', guild_id=guild_id))
 
-    def create_server(self, name, region, icon):
+    def create_guild(self, name, region, icon):
         payload = {
             'name': name,
             'icon': icon,
@@ -496,24 +587,39 @@ class HTTPClient:
 
         return self.request(Route('POST', '/guilds'), json=payload)
 
-    def edit_server(self, guild_id, **fields):
+    def edit_guild(self, guild_id, *, reason=None, **fields):
         valid_keys = ('name', 'region', 'icon', 'afk_timeout', 'owner_id',
-                      'afk_channel_id', 'splash', 'verification_level')
+                      'afk_channel_id', 'splash', 'verification_level',
+                      'system_channel_id', 'default_message_notifications',
+                      'description', 'explicit_content_filter', 'banner')
 
         payload = {
             k: v for k, v in fields.items() if k in valid_keys
         }
 
-        return self.request(Route('PATCH', '/guilds/{guild_id}', guild_id=guild_id), json=payload)
+        return self.request(Route('PATCH', '/guilds/{guild_id}', guild_id=guild_id), json=payload, reason=reason)
 
     def get_bans(self, guild_id):
         return self.request(Route('GET', '/guilds/{guild_id}/bans', guild_id=guild_id))
 
-    def prune_members(self, guild_id, days):
+    def get_ban(self, user_id, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/bans/{user_id}', guild_id=guild_id, user_id=user_id))
+
+    def get_vanity_code(self, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/vanity-url', guild_id=guild_id))
+
+    def change_vanity_code(self, guild_id, code, *, reason=None):
+        payload = {'code': code}
+        return self.request(Route('PATCH', '/guilds/{guild_id}/vanity-url', guild_id=guild_id), json=payload, reason=reason)
+
+    def get_member(self, guild_id, member_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/members/{member_id}', guild_id=guild_id, member_id=member_id))
+
+    def prune_members(self, guild_id, days, *, reason=None):
         params = {
             'days': days
         }
-        return self.request(Route('POST', '/guilds/{guild_id}/prune', guild_id=guild_id), params=params)
+        return self.request(Route('POST', '/guilds/{guild_id}/prune', guild_id=guild_id), params=params, reason=reason)
 
     def estimate_pruned_members(self, guild_id, days):
         params = {
@@ -521,28 +627,54 @@ class HTTPClient:
         }
         return self.request(Route('GET', '/guilds/{guild_id}/prune', guild_id=guild_id), params=params)
 
-    def create_custom_emoji(self, guild_id, name, image):
+    def get_all_custom_emojis(self, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/emojis', guild_id=guild_id))
+
+    def get_custom_emoji(self, guild_id, emoji_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/emojis/{emoji_id}', guild_id=guild_id, emoji_id=emoji_id))
+
+    def create_custom_emoji(self, guild_id, name, image, *, roles=None, reason=None):
         payload = {
             'name': name,
-            'image': image
+            'image': image,
+            'roles': roles or []
         }
 
         r = Route('POST', '/guilds/{guild_id}/emojis', guild_id=guild_id)
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def delete_custom_emoji(self, guild_id, emoji_id):
-        return self.request(Route('DELETE', '/guilds/{guild_id}/emojis/{emoji_id}', guild_id=guild_id, emoji_id=emoji_id))
+    def delete_custom_emoji(self, guild_id, emoji_id, *, reason=None):
+        r = Route('DELETE', '/guilds/{guild_id}/emojis/{emoji_id}', guild_id=guild_id, emoji_id=emoji_id)
+        return self.request(r, reason=reason)
 
-    def edit_custom_emoji(self, guild_id, emoji_id, *, name):
+    def edit_custom_emoji(self, guild_id, emoji_id, *, name, roles=None, reason=None):
         payload = {
-            'name': name
+            'name': name,
+            'roles': roles or []
         }
         r = Route('PATCH', '/guilds/{guild_id}/emojis/{emoji_id}', guild_id=guild_id, emoji_id=emoji_id)
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
+
+    def get_audit_logs(self, guild_id, limit=100, before=None, after=None, user_id=None, action_type=None):
+        params = {'limit': limit}
+        if before:
+            params['before'] = before
+        if after:
+            params['after'] = after
+        if user_id:
+            params['user_id'] = user_id
+        if action_type:
+            params['action_type'] = action_type
+
+        r = Route('GET', '/guilds/{guild_id}/audit-logs', guild_id=guild_id)
+        return self.request(r, params=params)
+
+    def get_widget(self, guild_id):
+        return self.request(Route('GET', '/guilds/{guild_id}/widget.json', guild_id=guild_id))
 
     # Invite management
 
-    def create_invite(self, channel_id, **options):
+    def create_invite(self, channel_id, *, reason=None, **options):
         r = Route('POST', '/channels/{channel_id}/invites', channel_id=channel_id)
         payload = {
             'max_age': options.get('max_age', 0),
@@ -551,10 +683,13 @@ class HTTPClient:
             'unique': options.get('unique', True)
         }
 
-        return self.request(r, json=payload)
+        return self.request(r, reason=reason, json=payload)
 
-    def get_invite(self, invite_id):
-        return self.request(Route('GET', '/invite/{invite_id}', invite_id=invite_id))
+    def get_invite(self, invite_id, *, with_counts=True):
+        params = {
+            'with_counts': int(with_counts)
+        }
+        return self.request(Route('GET', '/invite/{invite_id}', invite_id=invite_id), params=params)
 
     def invites_from(self, guild_id):
         return self.request(Route('GET', '/guilds/{guild_id}/invites', guild_id=guild_id))
@@ -562,48 +697,45 @@ class HTTPClient:
     def invites_from_channel(self, channel_id):
         return self.request(Route('GET', '/channels/{channel_id}/invites', channel_id=channel_id))
 
-    def accept_invite(self, invite_id):
-        return self.request(Route('POST', '/invite/{invite_id}', invite_id=invite_id))
-
-    def delete_invite(self, invite_id):
-        return self.request(Route('DELETE', '/invite/{invite_id}', invite_id=invite_id))
+    def delete_invite(self, invite_id, *, reason=None):
+        return self.request(Route('DELETE', '/invite/{invite_id}', invite_id=invite_id), reason=reason)
 
     # Role management
 
-    def edit_role(self, guild_id, role_id, **fields):
+    def edit_role(self, guild_id, role_id, *, reason=None, **fields):
         r = Route('PATCH', '/guilds/{guild_id}/roles/{role_id}', guild_id=guild_id, role_id=role_id)
         valid_keys = ('name', 'permissions', 'color', 'hoist', 'mentionable')
         payload = {
             k: v for k, v in fields.items() if k in valid_keys
         }
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def delete_role(self, guild_id, role_id):
+    def delete_role(self, guild_id, role_id, *, reason=None):
         r = Route('DELETE', '/guilds/{guild_id}/roles/{role_id}', guild_id=guild_id, role_id=role_id)
-        return self.request(r)
+        return self.request(r, reason=reason)
 
-    def replace_roles(self, user_id, guild_id, role_ids):
-        return self.edit_member(guild_id=guild_id, user_id=user_id, roles=role_ids)
+    def replace_roles(self, user_id, guild_id, role_ids, *, reason=None):
+        return self.edit_member(guild_id=guild_id, user_id=user_id, roles=role_ids, reason=reason)
 
-    def create_role(self, guild_id):
+    def create_role(self, guild_id, *, reason=None, **fields):
         r = Route('POST', '/guilds/{guild_id}/roles', guild_id=guild_id)
-        return self.request(r)
+        return self.request(r, json=fields, reason=reason)
 
-    def move_role_position(self, guild_id, positions):
+    def move_role_position(self, guild_id, positions, *, reason=None):
         r = Route('PATCH', '/guilds/{guild_id}/roles', guild_id=guild_id)
-        return self.request(r, json=positions)
+        return self.request(r, json=positions, reason=reason)
 
-    def add_role(self, guild_id, user_id, role_id):
+    def add_role(self, guild_id, user_id, role_id, *, reason=None):
         r = Route('PUT', '/guilds/{guild_id}/members/{user_id}/roles/{role_id}',
                   guild_id=guild_id, user_id=user_id, role_id=role_id)
-        return self.request(r)
+        return self.request(r, reason=reason)
 
-    def remove_role(self, guild_id, user_id, role_id):
+    def remove_role(self, guild_id, user_id, role_id, *, reason=None):
         r = Route('DELETE', '/guilds/{guild_id}/members/{user_id}/roles/{role_id}',
                   guild_id=guild_id, user_id=user_id, role_id=role_id)
-        return self.request(r)
+        return self.request(r, reason=reason)
 
-    def edit_channel_permissions(self, channel_id, target, allow, deny, type):
+    def edit_channel_permissions(self, channel_id, target, allow, deny, type, *, reason=None):
         payload = {
             'id': target,
             'allow': allow,
@@ -611,38 +743,82 @@ class HTTPClient:
             'type': type
         }
         r = Route('PUT', '/channels/{channel_id}/permissions/{target}', channel_id=channel_id, target=target)
-        return self.request(r, json=payload)
+        return self.request(r, json=payload, reason=reason)
 
-    def delete_channel_permissions(self, channel_id, target):
+    def delete_channel_permissions(self, channel_id, target, *, reason=None):
         r = Route('DELETE', '/channels/{channel_id}/permissions/{target}', channel_id=channel_id, target=target)
-        return self.request(r)
+        return self.request(r, reason=reason)
 
     # Voice management
 
-    def move_member(self, user_id, guild_id, channel_id):
-        return self.edit_member(guild_id=guild_id, user_id=user_id, channel_id=channel_id)
+    def move_member(self, user_id, guild_id, channel_id, *, reason=None):
+        return self.edit_member(guild_id=guild_id, user_id=user_id, channel_id=channel_id, reason=reason)
+
+    # Relationship related
+
+    def remove_relationship(self, user_id):
+        r = Route('DELETE', '/users/@me/relationships/{user_id}', user_id=user_id)
+        return self.request(r)
+
+    def add_relationship(self, user_id, type=None):
+        r = Route('PUT', '/users/@me/relationships/{user_id}', user_id=user_id)
+        payload = {}
+        if type is not None:
+            payload['type'] = type
+
+        return self.request(r, json=payload)
+
+    def send_friend_request(self, username, discriminator):
+        r = Route('POST', '/users/@me/relationships')
+        payload = {
+            'username': username,
+            'discriminator': int(discriminator)
+        }
+        return self.request(r, json=payload)
 
     # Misc
 
     def application_info(self):
         return self.request(Route('GET', '/oauth2/applications/@me'))
 
-    @asyncio.coroutine
-    def get_gateway(self):
+    async def get_gateway(self, *, encoding='json', v=6, zlib=True):
         try:
-            data = yield from self.request(Route('GET', '/gateway'))
-        except HTTPException as e:
-            raise GatewayNotFound() from e
-        return data.get('url') + '?encoding=json&v=6'
-
-    @asyncio.coroutine
-    def get_bot_gateway(self):
-        try:
-            data = yield from self.request(Route('GET', '/gateway/bot'))
-        except HTTPException as e:
-            raise GatewayNotFound() from e
+            data = await self.request(Route('GET', '/gateway'))
+        except HTTPException as exc:
+            raise GatewayNotFound() from exc
+        if zlib:
+            value = '{0}?encoding={1}&v={2}&compress=zlib-stream'
         else:
-            return data['shards'], data['url'] + '?encoding=json&v=6'
+            value = '{0}?encoding={1}&v={2}'
+        return value.format(data['url'], encoding, v)
+
+    async def get_bot_gateway(self, *, encoding='json', v=6, zlib=True):
+        try:
+            data = await self.request(Route('GET', '/gateway/bot'))
+        except HTTPException as exc:
+            raise GatewayNotFound() from exc
+
+        if zlib:
+            value = '{0}?encoding={1}&v={2}&compress=zlib-stream'
+        else:
+            value = '{0}?encoding={1}&v={2}'
+        return data['shards'], value.format(data['url'], encoding, v)
 
     def get_user_info(self, user_id):
         return self.request(Route('GET', '/users/{user_id}', user_id=user_id))
+
+    def get_user_profile(self, user_id):
+        return self.request(Route('GET', '/users/{user_id}/profile', user_id=user_id))
+
+    def get_mutual_friends(self, user_id):
+        return self.request(Route('GET', '/users/{user_id}/relationships', user_id=user_id))
+
+    def change_hypesquad_house(self, house_id):
+        payload = {'house_id': house_id}
+        return self.request(Route('POST', '/hypesquad/online'), json=payload)
+
+    def leave_hypesquad_house(self):
+        return self.request(Route('DELETE', '/hypesquad/online'))
+
+    def edit_settings(self, **payload):
+        return self.request(Route('PATCH', '/users/@me/settings'), json=payload)
